@@ -4,44 +4,74 @@ use warnings;
 use v5.10;
 
 use parent 'Plack::Component';
-use Plack::Util::Accessor qw(hook access app events safe);
+use Plack::Util::Accessor qw(hook events secret access safe);
 use Plack::Request;
-use Plack::Builder;
+use Plack::Middleware::HTTPExceptions;
 use Plack::Middleware::Access;
 use Carp qw(croak);
 use JSON qw(decode_json);
+use Scalar::Util qw(blessed);
 
-our $VERSION = '0.7';
+our $VERSION = '0.8';
 
-sub prepare_app {
-    my $self = shift;
+our @GITHUB_IPS = (
+    allow => "204.232.175.64/27",
+    allow => "192.30.252.0/22",
+);
 
-    if ( (ref $self->hook // '') ne 'ARRAY' ) {
-        $self->hook( [ $self->hook // () ] );
+sub github_webhook {
+    my $hook = shift;
+    if ( !ref $hook ) {
+        my $class = Plack::Util::load_class($hook, 'GitHub::WebHook');
+        $class = $class->new;
+        return sub { $class->call(@_) };
+    } elsif ( ref $hook eq 'HASH' ) {
+        my ($class, $args) = each %$hook;
+        $class = Plack::Util::load_class($class, 'GitHub::WebHook');
+        $class = $class->new( ref $args eq 'HASH' ? %$args : @$args );
+        return sub { $class->call(@_) };
+    } elsif ( blessed $hook and $hook->can('call') ) {
+        return sub { $hook->call(@_) };
+    } elsif ( (ref $hook // '') ne 'CODE') {
+        croak "hook must be a CODE or ARRAY of CODEs";
     }
-
-    foreach my $task (@{$self->hook}) {
-        if ( (ref $task // '') ne 'CODE') {
-            croak "hook must be a CODE or ARRAY of CODEs";
-        }
-    }
-
-    $self->access([
-        allow => "204.232.175.64/27",
-        allow => "192.30.252.0/22",
-        deny  => "all"
-    ]) unless $self->access;
-
-    $self->app( builder {
-        enable 'Access', rules => $self->access;
-        enable 'HTTPExceptions';
-        sub { $self->call_granted($_[0]) },
-    } );
+    $hook;
 }
 
-sub call {
-    my ($self, $env) = @_;
-    $self->app->($env);
+sub to_app {
+    my $self = shift;
+
+    my $hook = (ref $self->hook // '') eq 'ARRAY' 
+             ? $self->hook : [ $self->hook // () ];
+    $self->hook([ map { github_webhook($_) } @$hook ]);
+
+    my $app = Plack::Middleware::HTTPExceptions->wrap(
+        sub { $self->call_granted($_[0]) }
+    );
+
+    if ($self->secret) {
+        require Plack::Middleware::HubSignature;
+        $app = Plack::Middleware::HubSignature->wrap($app,
+            secret => $self->secret
+        );
+    }
+
+    $self->access('github') unless $self->access;
+    $self->access([]) if $self->access eq 'all';
+    my @rules = (@GITHUB_IPS, 'deny' => 'all');
+    if ( $self->access !~ /^github$/i ) {
+        @rules = ();
+        foreach (@{$self->access}) {
+            if (@rules and $rules[0] eq 'allow' and $_ =~ /^github$/i) {
+                push @rules, @GITHUB_IPS[1 .. $#GITHUB_IPS];
+            } else {
+                push @rules, $_;
+            }
+        }
+    }
+    $app = Plack::Middleware::Access->wrap( $app, rules => \@rules );
+
+    $app;
 }
 
 sub call_granted {
@@ -58,7 +88,8 @@ sub call_granted {
     my ($status, $message);
     
     if ( !$self->events or grep { $event eq $_ } @{$self->events} ) {
-        $payload = eval { decode_json $req->content };
+        $payload = $req->param('payload') || $req->content;
+        $payload = eval { decode_json $payload };
     }
 
     if (!$payload) {
@@ -151,29 +182,25 @@ Plack::App::GitHub::WebHook - GitHub WebHook receiver as Plack application
 
 =head1 SYNOPSIS
 
-=head2 Basic usage
-
     use Plack::App::GitHub::WebHook;
 
+    # Basic Usage
     Plack::App::GitHub::WebHook->new(
         hook => sub {
-            my ($payload, $event, $delivery, $logger) = @_;
+            my $payload = shift;
             ...
-        }
+        },
+        events => ['pull'],  # optional
+        secret => $secret,   # optional
+        access => 'github',  # default
     )->to_app;
 
-=head2 Multiple task hooks
-
-A hook can consist of multiple tasks, given by an array reference. The tasks
-are called one by one until a task returns a false value.
-
-    use Plack::App::GitHub::WebHook;
+    # Multiple hooks
     use IPC::Run3;
-
     Plack::App::GitHub::WebHook->new(
         hook => [
-            sub { $_[0]->{repository}{name} eq 'foo' }, # filter
-            sub { # action
+            sub { $_[0]->{repository}{name} eq 'foo' },
+            sub {
                 my ($payload, $event, $delivery, $logger) = @_;
                 run3 \@cmd, undef, $logger->{info}, $logger->{error}; 
             },
@@ -181,34 +208,170 @@ are called one by one until a task returns a false value.
         ]
     )->to_app;
 
-=head2 Access restriction    
+=head1 DESCRIPTION
 
-By default access is restricted to known GitHub WebHook IPs.
+This L<PSGI> application receives HTTP POST requests with body parameter
+C<payload> set to a JSON object. The default use case is to receive 
+L<GitHub WebHooks|http://developer.github.com/webhooks/>, for instance
+L<PushEvents|http://developer.github.com/v3/activity/events/types/#pushevent>.
 
-    Plack::App::GitHub::WebHook->new(
-        hook => sub { ... },
-        access => [
-            allow => "204.232.175.64/27",
-            allow => "192.30.252.0/22",
-            deny  => 'all'
-        ]
-    )->to_app;
+The response of a HTTP request to this application is one of:
 
-    # this is equivalent to
-    use Plack::Builder;
-    builder {
-        mount 'notify' => builder {
-            enable 'Access', rules => [
-                allow => "204.232.175.64/27",
-                allow => "192.30.252.0/22",
-                deny  => 'all'
-            ]
-            Plack::App::GitHub::WebHook->new(
-                hook => sub { ... }
-            );
-        }
+=over 4
+
+=item HTTP 403 Forbidden
+
+If access was not granted (for instance because it did not origin from GitHub).
+
+=item HTTP 405 Method Not Allowed
+
+If the request was no HTTP POST.
+
+=item HTTP 400 Bad Request
+
+If the payload was no well-formed JSON or the C<X-GitHub-Event> header did not
+match configured events.
+
+=item HTTP 200 OK
+
+Otherwise, if the hook was called and returned a true value.
+
+=item HTTP 202 Accepted
+
+Otherwise, if the hook was called and returned a false value.
+
+=item HTTP 500 Internal Server Error
+
+If a hook died with an exception, the error is returned as content body. Use
+configuration parameter C<safe> to disable HTTP 500 errors. 
+
+=back
+
+This module requires at least Perl 5.10.
+
+=head1 CONFIGURATION
+
+=over
+
+=item hook
+
+A hook can be any of a code reference, an object instance with method C<code>,
+a class name, or a class name mapped to parameters. You can also pass a list of
+hooks as array reference. Class names are prepended by L<GitHub::WebHook>
+unless prepended by C<+>.
+    
+    hook => sub {
+        my ($payload, $event, $delivery, $logger) = @_;
+        ...
+    }
+
+    hook => 'Foo'
+    hook => '+GitHub::WebHook::Foo'
+    hook => GitHub::WebHook::Foo->new
+
+    hook => { Bar => [ doz => 'baz' ] }
+    hook => GitHub::WebHook::Bar->new( doz => 'baz' )
+    
+Each hook gets passed the encoded payload, the type of webhook
+L<event|https://developer.github.com/webhooks/#events>, a unique delivery ID,
+and a L<logger object|/LOGGING>.  If the hook returns a true value, the next
+the hook is called or HTTP status code 200 is returned.  If a hook returns a
+false value (or if no hook was given), HTTP status code 202 is returned
+immediately.  Information can be passed from one hook to the next by modifying
+the payload. 
+
+=item events
+
+A list of L<event types|http://developer.github.com/v3/activity/events/types/>
+expected to be send with the C<X-GitHub-Event> header (e.g. C<['pull']>).
+
+=item secret
+
+Secret token set at GitHub Webhook setting to validate payload.  See
+L<https://developer.github.com/webhooks/securing/> for details. Requires
+L<Plack::Middleware::HubSignature>.
+
+=item access
+
+Access restrictions, as passed to L<Plack::Middleware::Access>. A recent list
+of official GitHub WebHook IPs is vailable at L<https://api.github.com/meta>.
+The default value
+
+    access => 'github'
+
+is a shortcut for these official IP ranges
+
+    access => [
+        allow => "204.232.175.64/27",
+        allow => "192.30.252.0/22",
+        deny  => 'all'
+    ]
+
+and
+
+    access => [
+        allow => 'github',
+        ...
+    ]
+
+is a shortcut for
+
+    access => [
+        allow => "204.232.175.64/27",
+        allow => "192.30.252.0/22",
+        ...
+    ]
+
+To disable access control via IP ranges use any of
+
+    access => 'all'
+    access => []
+
+=item safe
+
+Wrap all hooks in C<< eval { ... } >> blocks to catch exceptions.  Error
+messages are send to the PSGI error stream C<psgi.errors>.  A dying hook in
+safe mode is equivalent to a hook that returns a false value, so it will result
+in a HTTP 202 response.
+
+If you want errors to result in a HTTP 500 response, don't use this option but
+wrap the application in an eval block such as this:
+
+    sub {
+        eval { $app->(@_) } || do {
+            my $msg = $@ || 'Server Error';
+            [ 500, [ 'Content-Length' => length $msg ], [ $msg ] ];
+        };
     };
 
+
+=back
+
+=head1 LOGGING
+
+Each hook is passed a logging object as fourth parameter. It provides logging
+methods for each log level and a general log method:
+
+    sub sample_hook {
+        my ($payload, $event, $delivery, $log) = @_;
+
+        $log->debug('message');  $log->{debug}->('message');
+        $log->info('message');   $log->{info}->('message');
+        $log->warn('message');   $log->{warn}->('message');
+        $log->error('message');  $log->{error}->('message');
+        $log->fatal('message');  $log->{fatal}->('message');
+
+        $log->log( warn => 'message' );
+
+        run3 \@system_command, undef,
+            $log->{info},   # STDOUT to log level info
+            $log->{error};  # STDERR to log level error
+    }
+
+Trailing newlines on log messages are trimmed.
+
+=head1 EXAMPLES
+
 =head2 Synchronize with a GitHub repository
 
 The following application automatically pulls the master branch of a GitHub
@@ -247,122 +410,6 @@ repository into a local working directory.
         ],
     )->to_app;
 
-=head1 DESCRIPTION
-
-This L<PSGI> application receives HTTP POST requests with body parameter
-C<payload> set to a JSON object. The default use case is to receive 
-L<GitHub WebHooks|http://developer.github.com/webhooks/>, for instance
-L<PushEvents|http://developer.github.com/v3/activity/events/types/#pushevent>.
-
-The response of a HTTP request to this application is one of:
-
-=over 4
-
-=item HTTP 403 Forbidden
-
-If access was not granted (for instance because it did not origin from GitHub).
-
-=item HTTP 405 Method Not Allowed
-
-If the request was no HTTP POST.
-
-=item HTTP 400 Bad Request
-
-If the payload was no well-formed JSON or the C<X-GitHub-Event> header did not
-match configured events.
-
-=item HTTP 200 OK
-
-Otherwise, if the hook was called and returned a true value.
-
-=item HTTP 202 Accepted
-
-Otherwise, if the hook was called and returned a false value.
-
-=item HTTP 500 Internal Server Error
-
-If a hook died with an exception, the error is returned as content body.                
-
-=back
-
-This module requires at least Perl 5.10.
-
-=head1 CONFIGURATION
-
-=over 4
-
-=item hook
-
-A code reference or an array of code references with tasks that are executed on
-an incoming webhook.  Each task gets passed the encoded payload, the
-L<event|https://developer.github.com/webhooks/#events> and the unique delivery
-ID, and a L<logger object|/LOGGING>.  If the task returns a true value, next
-the task is called or HTTP status code 200 is returned.  Information can be
-passed from one task to the next by modifying the payload. 
-
-If a task returns a false value or if no task was given, HTTP status code 202
-is returned immediately. This mechanism can be used for conditional hooks or to
-detect hooks that were called successfully but failed to execute for some
-reason.
-
-=item safe
-
-Wrap all hook tasks in C<< eval { ... } >> blocks to catch exceptions.  Error
-messages are send to the PSGI error stream C<psgi.errors>.  A dying task in
-safe mode is equivalent to a task that returns a false value, so it will result
-in a HTTP 202 response.
-
-Plack::Middleware::HTTPExceptions
-
-If you want errors to result in a HTTP 500 response,
-wrap the application in an eval block such as this:
-
-    sub {
-        eval { $app->(@_) } || do {
-            my $msg = $@ || 'Server Error';
-            [ 500, [ 'Content-Length' => length $msg ], [ $msg ] ];
-        };
-    };
-
-=item access
-
-Access restrictions, as passed to L<Plack::Middleware::Access>. See SYNOPSIS
-for the default value. A recent list of official GitHub WebHook IPs is vailable
-at L<https://api.github.com/meta>. One should only set the access value on
-instantiation, or manually call C<prepare_app> after modification.
-
-=item events
-
-A list of L<event types|http://developer.github.com/v3/activity/events/types/>
-expected to be send with the C<X-GitHub-Event> header (e.g. C<['pull']>).
-
-=cut
-
-=back
-
-=head1 LOGGING
-
-Each hook is passed a logging object as fourth parameter. It provides logging
-methods for each log level and a general log method:
-
-    sub sample_hook {
-        my ($payload, $event, $delivery, $log) = @_;
-
-        $log->debug('message');  $log->{debug}->('message');
-        $log->info('message');   $log->{info}->('message');
-        $log->warn('message');   $log->{warn}->('message');
-        $log->error('message');  $log->{error}->('message');
-        $log->fatal('message');  $log->{fatal}->('message');
-
-        $log->log( warn => 'message' );
-
-        run3 \@system_command, undef,
-            $log->{info},   # STDOUT to log level info
-            $log->{error};  # STDERR to log level error
-    }
-
-Trailing newlines on log messages are trimmed.
-
 =head1 DEPLOYMENT
 
 Many deployment methods exist. An easy option might be to use Apache webserver
@@ -418,7 +465,8 @@ L<Net::GitHub> and L<Pithub> provide access to GitHub APIs.
 
 =item
 
-L<App::GitHubWebhooks2Ikachan> is an application that also receives GitHub WebHooks.
+L<Github::Hooks::Receiver> and L<App::GitHubWebhooks2Ikachan> are alternative
+application that receive GitHub WebHooks.
 
 =back
 
